"""
Library for building clusters on data.
"""
from typing import List, Dict, Any
from copy import copy
import itertools

import numpy as npy
from sklearn import cluster, preprocessing
import matplotlib.pyplot as plt

try:
    from plot_data.core import Scatter, Histogram, MultiplePlots, Tooltip, ParallelPlot, PointFamily, EdgeStyle, Axis, \
        PointStyle
    from plot_data.colors import BLUE, GREY, LIGHTGREY, Color
except ImportError:
    pass
from dessia_common.exports import XLSXWriter
from dessia_common.core import DessiaObject, DessiaFilter, FiltersList, templates


class HeterogeneousList(DessiaObject):
    """
    Base object for handling a list of DessiaObjects.

    :param dessia_objects:
        --------
        List of DessiaObjects to store in HeterogeneousList
    :type dessia_objects: `List[DessiaObject]`, `optional`, defaults to `None`

    :param name:
        --------
        Name of HeterogeneousList
    :type name: `str`, `optional`, defaults to `''`

    :Properties:
        * **common_attributes:** (`List[str]`)
            --------
            Common attributes of DessiaObjects contained in the current `HeterogeneousList`

        * **matrix:** (`List[List[float]]`, `n_samples x n_features`)
            --------
            Matrix of data computed by calling the `to_vector` method of all `dessia_objects`

    **Built-in methods**:
        * __init__
            >>> from dessia_common.core import HeterogeneousList
            >>> from dessia_common.models import all_cars_wi_feat
            >>> hlist = HeterogeneousList(all_cars_wi_feat, name="init")

        * __str__
            >>> print(HeterogeneousList(all_cars_wi_feat[:3], name='printed'))
            HeterogeneousList printed: 3 samples, 5 features
            |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
            -----------------------------------------------------------------------------------------------------------
            |               18.0  |             0.307  |             130.0  |            3504.0  |              12.0  |
            |               15.0  |              0.35  |             165.0  |            3693.0  |              11.5  |
            |               18.0  |             0.318  |             150.0  |            3436.0  |              11.0  |

        * __len__
            >>> len(HeterogeneousList(all_cars_wi_feat))
            returns len(all_cars_wi_feat)

        * __get_item__
            >>> HeterogeneousList(all_cars_wi_feat)[0]
            returns <dessia_common.tests.CarWithFeatures object at 'memory_address'>
            >>> HeterogeneousList(all_cars_wi_feat)[0:2]
            returns HeterogeneousList(all_cars_wi_feat[0:2])
            >>> HeterogeneousList(all_cars_wi_feat)[[0,5,6]]
            returns HeterogeneousList([all_cars_wi_feat[idx] for idx in [0,5,6]])
            >>> booleans_list = [True, False,..., True] of length len(all_cars_wi_feat)
            >>> HeterogeneousList(all_cars_wi_feat)[booleans_list]
            returns HeterogeneousList([car for car, boolean in zip(all_cars_wi_feat, booleans_list) if boolean])

        * __add__
            >>> HeterogeneousList(all_cars_wi_feat) + HeterogeneousList(all_cars_wi_feat)
            HeterogeneousList(all_cars_wi_feat + all_cars_wi_feat)
            >>> HeterogeneousList(all_cars_wi_feat) + HeterogeneousList()
            HeterogeneousList(all_cars_wi_feat)
            >>> HeterogeneousList(all_cars_wi_feat).extend(HeterogeneousList(all_cars_wi_feat))
            HeterogeneousList(all_cars_wi_feat + all_cars_wi_feat)
    """
    _standalone_in_db = True
    _vector_features = ["name", "common_attributes"]

    def __init__(self, dessia_objects: List[DessiaObject] = None, name: str = ''):
        if dessia_objects is None:
            dessia_objects = []
        self.dessia_objects = dessia_objects
        self._common_attributes = None
        self._matrix = None
        DessiaObject.__init__(self, name=name)

    def __getitem__(self, key: Any):
        """
        Is added in a further release (feat/clists_metrics)

        """
        if len(self) == 0:
            return []
        if isinstance(key, int):
            return self.pick_from_int(key)
        if isinstance(key, slice):
            return self.pick_from_slice(key)
        if isinstance(key, list):
            if len(key) == 0:
                return self.__class__()
            if isinstance(key[0], bool):
                if len(key) == self.__len__():
                    return self.pick_from_boolist(key)
                raise ValueError(f"Cannot index {self.__class__.__name__} object of len {self.__len__()} with a "
                                 f"list of boolean of len {len(key)}")
            if isinstance(key[0], int):
                return self.pick_from_boolist(self._indexlist_to_booleanlist(key))

        raise NotImplementedError(f"key of type {type(key)} with {type(key[0])} elements not implemented for "
                                  f"indexing HeterogeneousLists")

    def __add__(self, other: 'HeterogeneousList'):
        """
        Is added in a further release (feat/clists_metrics)

        """
        if self.__class__ != HeterogeneousList or other.__class__ != HeterogeneousList:
            raise TypeError("Addition only defined for HeterogeneousList. A specific __add__ method is required for "
                            f"{self.__class__}")

        sum_hlist = self.__class__(dessia_objects=self.dessia_objects + other.dessia_objects,
                                   name=self.name[:5] + '_+_' + other.name[:5])

        if all(item in self.common_attributes for item in other.common_attributes):
            sum_hlist._common_attributes = self.common_attributes
            if self._matrix is not None and other._matrix is not None:
                sum_hlist._matrix = self._matrix + other._matrix
        return sum_hlist

    def extend(self, other: 'HeterogeneousList'):
        """
        Update a HeterogeneousList by adding b values to it

        :param b: HeterogeneousList to add to the current HeterogeneousList
        :type b: HeterogeneousList

        :return: None

        Examples
        --------
        >>> from dessia_common.core import HeterogeneousList
        >>> from dessia_common.models import all_cars_wi_feat
        >>> HeterogeneousList(all_cars_wi_feat).extend(HeterogeneousList(all_cars_wi_feat))
        HeterogeneousList(all_cars_wi_feat + all_cars_wi_feat)
        """
        # Not "self.dessia_objects += other.dessia_objects" to take advantage of __add__ algorithm
        self.__dict__.update((self + other).__dict__)

    def pick_from_int(self, idx: int):
        return self.dessia_objects[idx]

    def pick_from_slice(self, key: slice):
        new_hlist = self.__class__(dessia_objects=self.dessia_objects[key], name=self.name)
        new_hlist._common_attributes = copy(self._common_attributes)
        new_hlist.dessia_objects = self.dessia_objects[key]
        if self._matrix is not None:
            new_hlist._matrix = self._matrix[key]
        # new_hlist.name += f"_{key.start if key.start is not None else 0}_{key.stop}")
        return new_hlist

    def _indexlist_to_booleanlist(self, index_list: List[int]):
        boolean_list = [False] * len(self)
        for idx in index_list:
            boolean_list[idx] = True
        return boolean_list

    def pick_from_boolist(self, key: List[bool]):
        new_hlist = self.__class__(dessia_objects=DessiaFilter.apply(self.dessia_objects, key), name=self.name)
        new_hlist._common_attributes = copy(self._common_attributes)
        if self._matrix is not None:
            new_hlist._matrix = DessiaFilter.apply(self._matrix, key)
        # new_hlist.name += "_list")
        return new_hlist

    def __str__(self):
        size_col_label = self._set_size_col_label()
        attr_name_len = []
        attr_space = []
        prefix = self._write_str_prefix()

        if self.__len__() == 0:
            return prefix

        string = ""
        string += self._print_titles(attr_space, attr_name_len, size_col_label)
        string += "\n" + "-" * len(string)

        string += self._print_objects_slice(slice(0, 5), attr_space, attr_name_len,
                                            self._set_label_space(size_col_label))

        undispl_len = len(self) - 10
        string += (f"\n+ {undispl_len} undisplayed object" + "s" * (min([undispl_len, 2]) - 1) + "..."
                   if len(self) > 10 else '')

        string += self._print_objects_slice(slice(-5, len(self)), attr_space, attr_name_len,
                                            self._set_label_space(size_col_label))
        return prefix + "\n" + string + "\n"

    def _print_objects_slice(self, key: slice, attr_space: int, attr_name_len: int, label_space: int):
        string = ""
        for dessia_object in self.dessia_objects[key]:
            string += "\n" + " " * label_space
            string += self._print_objects(dessia_object, attr_space, attr_name_len)
        return string

    def _set_size_col_label(self):
        return 0

    def _set_label_space(self, size_col_label: int):
        if size_col_label:
            return 2 * size_col_label - 1
        return 0

    def _write_str_prefix(self):
        prefix = f"{self.__class__.__name__} {self.name if self.name != '' else hex(id(self))}: "
        prefix += f"{len(self)} samples, {len(self.common_attributes)} features"
        return prefix

    def _print_titles(self, attr_space: int, attr_name_len: int, size_col_label: int):
        string = ""
        if size_col_label:
            string += "|" + " " * (size_col_label - 1) + "n°" + " " * (size_col_label - 1)
        for idx, attr in enumerate(self.common_attributes):
            end_bar = ""
            if idx == len(self.common_attributes) - 1:
                end_bar = "|"
            # attribute
            attr_space.append(len(attr) + 6)
            name_attr = " " * 3 + f"{attr.capitalize()}" + " " * 3
            attr_name_len.append(len(name_attr))
            string += "|" + name_attr + end_bar
        return string

    def _print_objects(self, dessia_object: DessiaObject, attr_space: int, attr_name_len: int):
        string = ""
        for idx, attr in enumerate(self.common_attributes):
            end_bar = ""
            if idx == len(self.common_attributes) - 1:
                end_bar = "|"

            # attribute
            string += "|" + " " * (attr_space[idx] - len(str(getattr(dessia_object, attr))) - 1)
            string += f"{getattr(dessia_object, attr)}"[:attr_name_len[idx] - 3]
            if len(str(getattr(dessia_object, attr))) > attr_name_len[idx] - 3:
                string += "..."
            else:
                string += " "
            string += end_bar
        return string

    def __len__(self):
        """
        Is added in a further release (feat/clists_metrics)

        """
        return len(self.dessia_objects)

    def get_attribute_values(self, attribute: str):
        """
        Get a list of all values of dessia_objects of an attribute given by name

        :param attribute: Attribute to get all values
        :type attribute: str

        :return: A list of all values of the specified attribute of dessia_objects
        :rtype: List[Any]

        Examples
        --------
        >>> from dessia_common.core import HeterogeneousList
        >>> from dessia_common.models import all_cars_wi_feat
        >>> HeterogeneousList(all_cars_wi_feat[:10]).get_attribute_values("weight")
        [3504.0, 3693.0, 3436.0, 3433.0, 3449.0, 4341.0, 4354.0, 4312.0, 4425.0, 3850.0]
        """
        return [getattr(dessia_object, attribute) for dessia_object in self.dessia_objects]

    def get_column_values(self, index: int):
        """
        Get a list of all values of dessia_objects for an attribute given by its index common_attributes

        :param index: Index in common_attributes to get all values of dessia_objects
        :type index: int

        :return: A list of all values of the specified attribute of dessia_objects
        :rtype: List[float]

        Examples
        --------
        >>> from dessia_common.core import HeterogeneousList
        >>> from dessia_common.models import all_cars_wi_feat
        >>> HeterogeneousList(all_cars_wi_feat[:10]).get_column_values(2)
        [130.0, 165.0, 150.0, 150.0, 140.0, 198.0, 220.0, 215.0, 225.0, 190.0]
        """
        return [row[index] for row in self.matrix]

    def sort(self, key: Any, ascend: bool = True):  # TODO : Replace numpy with faster algorithms
        """
        Sort the current HeterogeneousList along the given key.

        :param key:
            --------
            The parameter on which to sort the HeterogeneousList. Can be an attribute or its index in \
                `common_attributes`
        :type key: `int` or `str`

        :param ascend:
            --------
            Whether to sort the HeterogeneousList in ascending (`True`) or descending (`False`) order
        :type key: `bool`, defaults to `True`

        :return: None

        Examples
        --------
        >>> from dessia_common.core import HeterogeneousList
        >>> from dessia_common.models import all_cars_wi_feat
        >>> example_list = HeterogeneousList(all_cars_wi_feat[:3], "sort_example")
        >>> example_list.sort("mpg", False)
        >>> print(example_list)
        HeterogeneousList sort_example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               18.0  |             0.318  |             150.0  |            3436.0  |              11.0  |
        |               18.0  |             0.307  |             130.0  |            3504.0  |              12.0  |
        |               15.0  |              0.35  |             165.0  |            3693.0  |              11.5  |
        >>> example_list.sort(2, True)
        >>> print(example_list)
        HeterogeneousList sort_example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               18.0  |             0.307  |             130.0  |            3504.0  |              12.0  |
        |               18.0  |             0.318  |             150.0  |            3436.0  |              11.0  |
        |               15.0  |              0.35  |             165.0  |            3693.0  |              11.5  |
        """
        if len(self) != 0:
            if isinstance(key, int):
                sort_indexes = npy.argsort(self.get_column_values(key))
            elif isinstance(key, str):
                sort_indexes = npy.argsort(self.get_attribute_values(key))
            self.dessia_objects = [self.dessia_objects[idx] for idx in (sort_indexes if ascend else sort_indexes[::-1])]
            if self._matrix is not None:
                self._matrix = [self._matrix[idx] for idx in (sort_indexes if ascend else sort_indexes[::-1])]

    @property
    def common_attributes(self):
        if self._common_attributes is None:
            if len(self) == 0:
                return []

            all_class = []
            one_instance = []
            for dessia_object in self.dessia_objects:
                if dessia_object.__class__ not in all_class:
                    all_class.append(dessia_object.__class__)
                    one_instance.append(dessia_object)

            all_attributes = sum((instance.vector_features() for instance in one_instance), [])
            set_attributes = set.intersection(*(set(instance.vector_features()) for instance in one_instance))

            # Keep order
            self._common_attributes = []
            for attr in all_attributes:
                if attr in set_attributes:
                    self._common_attributes.append(attr)
                    set_attributes.remove(attr)

        return self._common_attributes

    @property
    def matrix(self):
        if self._matrix is None:
            matrix = []
            for dessia_object in self.dessia_objects:
                temp_row = dessia_object.to_vector()
                vector_features = dessia_object.vector_features()
                matrix.append(list(temp_row[vector_features.index(attr)] for attr in self.common_attributes))
            self._matrix = matrix
        return self._matrix

    def filtering(self, filters_list: FiltersList):
        """
        Filter a HeterogeneousList given a FiltersList.
        Method filtering apply a FiltersList to the current HeterogeneousList.

        :param filters_list:
            FiltersList to apply on current HeterogeneousList
        :type filters_list: FiltersList

        :return: The filtered HeterogeneousList
        :rtype: HeterogeneousList

        Examples
        --------
        >>> from dessia_common.core import HeterogeneousList, DessiaFilter
        >>> from dessia_common.models import all_cars_wi_feat
        >>> filters = [DessiaFilter('weight', '<=', 1650.), DessiaFilter('mpg', '>=', 45.)]
        >>> filters_list = FiltersList(filters, "xor")
        >>> example_list = HeterogeneousList(all_cars_wi_feat, name="example")
        >>> filtered_list = example_list.filtering(filters_list)
        >>> print(filtered_list)
        HeterogeneousList example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               35.0  |             0.072  |              69.0  |            1613.0  |              18.0  |
        |               31.0  |             0.076  |              52.0  |            1649.0  |              16.5  |
        |               46.6  |             0.086  |              65.0  |            2110.0  |              17.9  |
        """
        booleans_index = filters_list.get_booleans_index(self.dessia_objects)
        return self.pick_from_boolist(booleans_index)

    def singular_values(self):
        """
        Computes the Singular Values Decomposition (SVD) of self.matrix.
        SVD factorizes self.matrix into two unitary matrices `U` and `Vh`, and a 1-D array `s` of singular values \
            (real, non-negative) such that ``a = U @ S @ Vh``, where S is diagonal such as `s1 > s2 >...> sn`.

        SVD gives indications on the dimensionality of a given matrix thanks to the normalized singular values: they \
            are stored in descending order and their sum is equal to 1. Thus, one can set a threshold value, e.g. \
                `0.95`, and keep only the `r` first normalized singular values which sum is greater than the threshold.

        `r` is the rank of the matrix and gives a good indication on the real dimensionality of the data contained in \
            the current HeterogeneousList. `r` is often much smaller than the current dimension of the studied data.
        This indicates that the used features can be combined into less new features, which do not necessarily \
            make sense for engineers.

        More informations: https://en.wikipedia.org/wiki/Singular_value_decomposition

        :return:
            **normalized_singular_values**: list of normalized singular values
            **singular_points**: list of points to plot in dimensionality plot. Does not add any information.
        :rtype: Tuple[List[float], List[Dict[str, float]]]

        """
        scaled_data = HeterogeneousList.scale_data(npy.array(self.matrix) - npy.mean(self.matrix, axis=0))
        _, singular_values, _ = npy.linalg.svd(npy.array(scaled_data).T, full_matrices=False)
        normalized_singular_values = singular_values / npy.sum(singular_values)

        singular_points = []
        for idx, value in enumerate(normalized_singular_values):
            # TODO (plot_data log_scale 0)
            singular_points.append({'Index of reduced basis vector': idx + 1,
                                    'Singular value': (value if value != 0. else 1e-16)})
        return normalized_singular_values, singular_points

    @staticmethod
    def scale_data(data_matrix: List[List[float]]):
        scaled_matrix = preprocessing.StandardScaler().fit_transform(data_matrix)
        return [list(map(float, row.tolist())) for row in scaled_matrix]

    def plot_data(self):
        """
        Plot data method.
        Plot a standard scatter matrix of all attributes in common_attributes and a dimensionality plot.
        """
        # Plot a correlation matrix : To develop
        # correlation_matrix = []
        # List datadict
        data_list = self._plot_data_list()
        # Dimensionality plot
        dimensionality_plot = self._plot_dimensionality()
        # Scattermatrix
        scatter_matrix = self._build_multiplot(data_list, self._tooltip_attributes(), axis=dimensionality_plot.axis,
                                               point_style=dimensionality_plot.point_style)
        # Parallel plot
        parallel_plot = self.parallel_plot(data_list)

        return [parallel_plot, scatter_matrix, dimensionality_plot]

    def _build_multiplot(self, data_list: List[Dict[str, float]], tooltip: List[str], **kwargs: Dict[str, Any]):
        subplots = []
        for line in self.common_attributes:
            for col in self.common_attributes:
                if line == col:
                    unic_values = set((getattr(dobject, line) for dobject in self.dessia_objects))
                    if len(unic_values) == 1:  # TODO (plot_data linspace axis between two same values)
                        subplots.append(Scatter(x_variable=line, y_variable=col))
                    else:
                        subplots.append(Histogram(x_variable=line))
                else:
                    subplots.append(Scatter(x_variable=line, y_variable=col, tooltip=Tooltip(tooltip), **kwargs))

        scatter_matrix = MultiplePlots(plots=subplots, elements=data_list, point_families=self._point_families(),
                                       initial_view_on=True)
        return scatter_matrix

    def parallel_plot(self, data_list: List[Dict[str, float]]):
        return ParallelPlot(elements=data_list, axes=self._parallel_plot_attr(), disposition='vertical')

    def _tooltip_attributes(self):
        return self.common_attributes

    def _plot_data_list(self):
        plot_data_list = []
        for row, _ in enumerate(self.dessia_objects):
            plot_data_list.append({attr: self.matrix[row][col] for col, attr in enumerate(self.common_attributes)})
        return plot_data_list

    def _point_families(self):
        return [PointFamily(BLUE, list(range(len(self))))]

    def _parallel_plot_attr(self):
        # TODO: Put it in plot_data
        (sorted_r2, sorted_association), constant_attributes = self._get_correlations()
        attr_series = self._get_attribute_trios(sorted_r2, sorted_association)
        return constant_attributes + self._trios_list_to_parallel_axes(attr_series)

    def _get_correlations(self):
        r2_scores = []
        association_list = []
        constant_attributes = []
        for idx, attr1 in enumerate(self.common_attributes):
            if len(set(self.get_attribute_values(attr1))) == 1:
                constant_attributes.append(attr1)
                continue
            for _, attr2 in enumerate(self.common_attributes[idx:]):
                if len(set(self.get_attribute_values(attr2))) == 1:
                    constant_attributes.append(attr2)
                    continue
                if attr1 != attr2:
                    correlation_matrix = npy.corrcoef(self.get_attribute_values(attr1),
                                                      self.get_attribute_values(attr2))
                    correlation_xy = correlation_matrix[0, 1]
                    r2_scores.append(correlation_xy**2)
                    association_list.append([attr1, attr2])
        # Returns list of list of associated attributes sorted along their R2 score and constant attributes
        return map(list, zip(*sorted(zip(r2_scores, association_list))[::-1])), list(set(constant_attributes))

    def _get_attribute_trios(self, sorted_r2, sorted_association):
        attr_series = []
        picked_attr = set()
        set_association = set(sum(sorted_association, []))
        while len(picked_attr) != len(set_association):
            first_association = sorted_association[0]
            attr_series.append(first_association)
            picked_attr.update(set(first_association))
            del sorted_r2[0], sorted_association[0]

            for idx, _ in enumerate(sorted_r2):
                if any(item in first_association for item in sorted_association[idx]):
                    attr_series[-1] += sorted_association[idx]
                    picked_attr.update(set(sorted_association[idx]))
                    del sorted_r2[idx], sorted_association[idx]
                    break
        return attr_series

    def _trios_list_to_parallel_axes(self, attribute_series):
        ordered_attr = []
        for attribute_serie in attribute_series:
            if not any(item in attribute_serie for item in ordered_attr):
                ordered_attr += self._new_attributes_trio(attribute_serie)
            else:
                ordered_attr = self._new_sided_attribute(ordered_attr, attribute_serie)
        return ordered_attr

    def _new_attributes_trio(self, attribute_serie):
        mid_index = [attribute_serie.count(attr) for attr in attribute_serie].index(2)
        mid_attr = attribute_serie[mid_index]
        remaining_attr = iter(set(attribute_serie).difference({mid_attr}))
        return [next(remaining_attr), mid_attr, next(remaining_attr)]

    def _new_sided_attribute(self, ordered_attr, attribute_serie):
        for side in [0, -1]:
            if ordered_attr[side] in attribute_serie:
                nb_instances = attribute_serie.count(ordered_attr[side])
                for ieme_instance in range(nb_instances):
                    idx_in_serie = (ieme_instance) * 2 + attribute_serie[ieme_instance * 2:].index(ordered_attr[side])
                    # 1 if idx_in_serie = 0, 0 if idx_in_serie = 1, 3 if idx_in_serie = 2, 2 if idx_in_serie = 3
                    idx_attr_to_add = idx_in_serie + 1 - 2 * (idx_in_serie % 2)
                    added_attr = []
                    if attribute_serie[idx_attr_to_add] not in ordered_attr:
                        added_attr = [attribute_serie[idx_attr_to_add]]
                        ordered_attr = (side + 1) * added_attr + ordered_attr + (-1 * side) * added_attr
                        break
        return ordered_attr

    def _plot_dimensionality(self):
        _, singular_points = self.singular_values()

        axis_style = EdgeStyle(line_width=0.5, color_stroke=GREY)
        axis = Axis(nb_points_x=len(singular_points), nb_points_y=len(singular_points), axis_style=axis_style)
        point_style = PointStyle(color_fill=BLUE, color_stroke=BLUE, stroke_width=0.1, size=2, shape='circle')

        dimensionality_plot = Scatter(elements=singular_points, x_variable='Index of reduced basis vector',
                                      y_variable='Singular value', log_scale_y=True, axis=axis, point_style=point_style)
        return dimensionality_plot

    def to_markdown(self):  # TODO: Custom this markdown
        """
        Render a markdown of the object output type: string
        """
        return templates.heterogeneouslist_markdown_template.substitute(name=self.name, class_=self.__class__.__name__)

    @staticmethod
    def _check_costs(len_data: int, costs: List[List[float]]):
        if len(costs) != len_data:
            if len(costs[0]) == len_data:
                return list(map(list, zip(*costs)))
            raise ValueError(f"costs is length {len(costs)} and the matching HeterogeneousList is length {len_data}. " +
                             "They should be the same length.")
        return costs

    @staticmethod
    def pareto_indexes(costs: List[List[float]]):
        """
        Find the pareto-efficient points

        :return: A (n_points, ) boolean list, indicating whether each point is Pareto efficient
        """
        is_efficient = npy.ones(len(costs), dtype=bool)
        # costs_array = npy.array(costs)
        costs_array = (costs - npy.mean(costs, axis=0)) / npy.std(costs, axis=0)
        for index, cost in enumerate(costs_array):
            if is_efficient[index]:
                # Keep any point with a lower cost
                is_efficient[is_efficient] = npy.any(costs_array[is_efficient] < cost, axis=1)
                # And keep self
                is_efficient[index] = True
        return is_efficient.tolist()

    def pareto_points(self, costs: List[List[float]]):
        """
        Find the pareto-efficient points

        :param costs:
            -----------
            costs on which the pareto points are computed
        :type costs: `List[List[float]]`, `n_samples x n_features`

        :return: a HeterogeneousList containing the selected points
        :rtype: HeterogeneousList
        """
        checked_costs = HeterogeneousList._check_costs(len(self), costs)
        return self[self.__class__.pareto_indexes(checked_costs)]

    def pareto_sheets(self, costs: List[List[float]], nb_sheets: int = 1):
        """
        Get successive pareto sheets (i.e. optimal points in a DOE for pre-computed costs).

        :param costs:
            Pre-computed costs of `len(self)`. Can be multi-dimensional.
        :type costs: `List[List[float]]`, `n_samples x n_costs` or `n_costs x n_samples`

        :param nb_sheets:
            Number of pareto sheets to pick
        :type nb_sheets: `int`, `optional`, default to `1`

        :return: The successive pareto sheets and not selected elements
        :rtype: `List[HeterogeneousList]`, `HeterogeneousList`
        """
        checked_costs = HeterogeneousList._check_costs(len(self), costs)
        non_optimal_costs = checked_costs[:]
        non_optimal_points = self.dessia_objects[:]
        pareto_sheets = []
        for idx in range(nb_sheets):
            pareto_sheet = HeterogeneousList.pareto_indexes(non_optimal_costs)
            pareto_sheets.append(HeterogeneousList(list(itertools.compress(non_optimal_points, pareto_sheet)),
                                                   self.name + f'_pareto_{idx}'))
            non_optimal_points = list(itertools.compress(non_optimal_points, map(lambda x: not x, pareto_sheet)))
            non_optimal_costs = list(itertools.compress(non_optimal_costs, map(lambda x: not x, pareto_sheet)))
        return pareto_sheets, HeterogeneousList(non_optimal_points, self.name)

    @staticmethod
    def pareto_frontiers(len_data: int, costs: List[List[float]]):
        # Experimental
        checked_costs = HeterogeneousList._check_costs(len_data, costs)
        pareto_indexes = HeterogeneousList.pareto_indexes(checked_costs)
        pareto_costs = npy.array(list(itertools.compress(checked_costs, pareto_indexes)))

        array_costs = npy.array(checked_costs)
        super_mini = npy.min(array_costs, axis=0)
        pareto_frontiers = []
        for x_dim in range(pareto_costs.shape[1]):
            for y_dim in range(pareto_costs.shape[1]):
                if x_dim != y_dim:
                    frontier_2d = HeterogeneousList._pareto_frontier_2d(x_dim, y_dim, pareto_costs,
                                                                        npy.max(array_costs[:, x_dim]), super_mini)
                    pareto_frontiers.append(frontier_2d)

        return pareto_frontiers

    @staticmethod
    def _pareto_frontier_2d(x_dim: int, y_dim: int, pareto_costs: List[List[float]], max_x_dim: float,
                            super_mini: List[float]):
        # Experimental
        minidx = npy.argmin(pareto_costs[:, y_dim])
        x_coord = pareto_costs[minidx, x_dim]
        y_coord = pareto_costs[minidx, y_dim]

        new_pareto = pareto_costs[x_coord - pareto_costs[:, x_dim] != 0., :]

        dir_coeffs = (y_coord - new_pareto[:, y_dim]) / (x_coord - new_pareto[:, x_dim])
        dir_coeffs[x_coord == new_pareto[:, x_dim]] = npy.max(dir_coeffs[x_coord != new_pareto[:, x_dim]])

        offsets = y_coord - dir_coeffs * x_coord
        approx_super_mini = dir_coeffs * super_mini[x_dim] + offsets
        chosen_line = npy.argmin(npy.absolute(approx_super_mini - super_mini[y_dim]))

        frontier_2d = npy.array([[super_mini[x_dim], max_x_dim], [approx_super_mini[chosen_line], max_x_dim *
                                                                  dir_coeffs[chosen_line] + offsets[chosen_line]]]).T
        return frontier_2d


class CategorizedList(HeterogeneousList):
    """
    Base object for handling a categorized (clustered) list of DessiaObjects.

    **CategorizedList should be instantiated with** `from_...` **methods.**

    **Do not use** `__init__` **to instantiate a CategorizedList.**

    :param dessia_objects:
        --------
        List of DessiaObjects to store in CategorizedList
    :type dessia_objects: `List[DessiaObject]`, `optional`, defaults to `None`

    :param labels:
        --------
        Labels of DessiaObjects' cluster stored in CategorizedList
    :type labels: `List[int]`, `optional`, defaults to `None`

    :param name:
        --------
        Name of CategorizedList
    :type name: `str`, `optional`, defaults to `""`

    :Properties:
        * **common_attributes:** (`List[str]`)
            --------
            Common attributes of DessiaObjects contained in the current `CategorizedList`
        * **matrix:** (`List[List[float]]`, `n_samples x n_features`)
            --------
            Matrix of data computed by calling the to_vector method of all dessia_objects
        * **n_cluster:** (`int`)
            --------
            Number of clusters in dessia_objects

    **Built-in methods**: See :func:`~HeterogeneousList`
    """
    _allowed_methods = ['from_agglomerative_clustering', 'from_kmeans', 'from_dbscan', 'from_pareto_sheets']

    def __init__(self, dessia_objects: List[DessiaObject] = None, labels: List[int] = None, name: str = ''):
        HeterogeneousList.__init__(self, dessia_objects=dessia_objects, name=name)
        if labels is None:
            labels = [0] * len(self)
        self.labels = labels
        self._n_clusters = None

    @property
    def n_clusters(self):
        if self._n_clusters is None:
            unic_labels = set(self.labels)
            unic_labels.discard(-1)
            self._n_clusters = len(unic_labels)
        return self._n_clusters

    def to_xlsx_stream(self, stream):
        """
        Exports the object to an XLSX to a given stream
        """
        if not isinstance(self.dessia_objects[0], HeterogeneousList):
            writer = XLSXWriter(self.clustered_sublists())
        else:
            writer = XLSXWriter(self)
        writer.save_to_stream(stream)

    def pick_from_slice(self, key: slice):
        new_hlist = HeterogeneousList.pick_from_slice(self, key)
        new_hlist.labels = self.labels[key]
        # new_hlist.name += f"_{key.start if key.start is not None else 0}_{key.stop}")
        return new_hlist

    def pick_from_boolist(self, key: List[bool]):
        new_hlist = HeterogeneousList.pick_from_boolist(self, key)
        new_hlist.labels = DessiaFilter.apply(self.labels, key)
        # new_hlist.name += "_list")
        return new_hlist

    def _print_objects_slice(self, key: slice, attr_space: int, attr_name_len: int, label_space: str):
        string = ""
        for label, dessia_object in zip(self.labels[key], self.dessia_objects[key]):
            string += "\n"
            space = label_space - len(str(label))
            string += "|" + " " * space + f"{label}" + " "
            string += self._print_objects(dessia_object, attr_space, attr_name_len)
        return string

    def _write_str_prefix(self):
        prefix = f"{self.__class__.__name__} {self.name if self.name != '' else hex(id(self))}: "
        prefix += (f"{len(self)} samples, {len(self.common_attributes)} features, {self.n_clusters} clusters")
        return prefix

    def _set_size_col_label(self):
        return 4

    def clustered_sublists(self):
        """
        Split a CategorizedList of labelled DessiaObjects into a CategorizedList of labelled HeterogeneousLists.

        :return: A CategorizedList of length n_cluster that store each cluster in a HeterogeneousList. Labels are \
            the labels of each cluster, i.e. stored HeterogeneousList
        :rtype: CategorizedList[HeterogeneousList]

        Examples
        --------
        >>> from dessia_common.core import HeterogeneousList
        >>> from dessia_common.models import all_cars_wi_feat
        >>> hlist = HeterogeneousList(all_cars_wi_feat, name="cars")
        >>> clist = CategorizedList.from_agglomerative_clustering(hlist, n_clusters=10, name="ex")
        >>> split_clist = clist.clustered_sublists()
        >>> print(split_clist[:3])
        CategorizedList ex_split: 3 samples, 2 features, 3 clusters
        |   n°   |   Name   |   Common_attributes   |
        ---------------------------------------------
        |      0 |     ex_0 |['mpg', 'displacemen...|
        |      1 |     ex_1 |['mpg', 'displacemen...|
        |      2 |     ex_2 |['mpg', 'displacemen...|
        >>> print(split_clist[3][:3])
        HeterogeneousList ex_3: 3 samples, 5 features
        |   Mpg   |   Displacement   |   Horsepower   |   Weight   |   Acceleration   |
        -------------------------------------------------------------------------------
        |    21.0 |              0.2 |           85.0 |     2587.0 |             16.0 |
        |    25.0 |             0.11 |           87.0 |     2672.0 |             17.5 |
        |    21.0 |            0.199 |           90.0 |     2648.0 |             15.0 |
        """
        sublists = []
        label_tags = sorted(list(map(str, set(self.labels).difference({-1}))))
        unic_labels = list(set(self.labels))
        for _ in range(self.n_clusters):
            sublists.append([])
        if -1 in self.labels:
            sublists.append([])
            label_tags.append("outliers")

        for idx, label in enumerate(self.labels):
            sublists[unic_labels.index(label)].append(self.dessia_objects[idx])

        new_dessia_objects = [HeterogeneousList(dessia_objects=sublist, name=self.name + f"_{label_tag}")
                              for label_tag, sublist in zip(label_tags, sublists)]

        return CategorizedList(new_dessia_objects,
                               list(set(self.labels).difference({-1})) + ([-1] if -1 in self.labels else []),
                               name=self.name + "_split")

    def _merge_sublists(self):
        merged_hlists = self.dessia_objects[0][:]
        merged_labels = [self.labels[0]] * len(merged_hlists)
        for dobject, label in zip(self.dessia_objects[1:], self.labels[1:]):
            merged_hlists.extend(dobject)
            merged_labels.extend([label] * len(dobject))
        plotted_clist = self.__class__(dessia_objects=merged_hlists.dessia_objects, labels=merged_labels)
        return plotted_clist

    def _tooltip_attributes(self):
        return self.common_attributes + ["Cluster Label"]

    def plot_data(self):
        if isinstance(self.dessia_objects[0], HeterogeneousList):
            plotted_clist = self._merge_sublists()
            return plotted_clist.plot_data()
        return HeterogeneousList.plot_data(self)

    def _plot_data_list(self):
        _plot_data_list = []
        for row, label in enumerate(self.labels):
            _plot_data_list.append({attr: self.matrix[row][col] for col, attr in enumerate(self.common_attributes)})
            _plot_data_list[-1]["Cluster Label"] = label
            # (label if label != -1 else "Excluded") plot_data "Excluded" -> NaN
        return _plot_data_list

    def _point_families(self):
        colormap = plt.cm.get_cmap('hsv', self.n_clusters + 1)(range(self.n_clusters + 1))
        point_families = []
        for i_cluster in range(self.n_clusters):
            color = Color(colormap[i_cluster][0], colormap[i_cluster][1], colormap[i_cluster][2])
            points_index = list(map(int, npy.where(npy.array(self.labels) == i_cluster)[0].tolist()))
            point_families.append(PointFamily(color, points_index, name="Cluster " + str(i_cluster)))

        if -1 in self.labels:
            color = LIGHTGREY
            points_index = list(map(int, npy.where(npy.array(self.labels) == -1)[0].tolist()))
            point_families.append(PointFamily(color, points_index, name="Excluded"))
        return point_families

    @classmethod
    def from_agglomerative_clustering(cls, data: HeterogeneousList, n_clusters: int = 2,
                                      affinity: str = 'euclidean', linkage: str = 'ward',
                                      distance_threshold: float = None, scaling: bool = False, name: str = ""):
        """
        Hierarchical clustering is a general family of clustering algorithms that
        build nested clusters by merging or splitting them successively.
        This hierarchy of clusters is represented as a tree (or dendrogram).
        The root of the tree is the unique cluster that gathers all the samples,
        the leaves being the clusters with only one sample. See the Wikipedia page
        for more details.

        The AgglomerativeClustering object performs a hierarchical clustering using
        a bottom up approach: each observation starts in its own cluster, and clusters
        are successively merged together. The linkage criteria determines the metric
        used for the merge strategy: Ward minimizes the sum of squared differences within all clusters.

        It is a variance-minimizing approach and in this sense is similar to the
        k-means objective function but tackled with an agglomerative hierarchical approach.
        Maximum or complete linkage minimizes the maximum distance between observations of pairs of clusters.
        Average linkage minimizes the average of the distances between all observations of pairs of clusters.
        Single linkage minimizes the distance between the closest observations of pairs of clusters.
        AgglomerativeClustering can also scale to large number of samples when it is used
        jointly with a connectivity matrix, but is computationally expensive when no connectivity
        constraints are added between samples: it considers at each step all the possible merges.

        See more : https://scikit-learn.org/stable/modules/clustering.html#hierarchical-clustering

        :param data: The future clustered data.
        :type data: List[DessiaObject]

        :param n_clusters:
            -------
            Number of wished clusters.

            Must be `None` if `distance_threshold` is not `None`
        :type n_clusters: `int`, `optional`, defaults to `2`

        :param affinity:
            -------
            Metric used to compute the linkage.
            Can be one of `['euclidean', 'l1', 'l2', 'manhattan', 'cosine', or 'precomputed']`.

            If linkage is `'ward'`, only `'euclidean'` is accepted.

            If `'precomputed'`, a distance matrix (instead of a similarity matrix) is needed as input for the \
                fit method.
        :type affinity: `str`, `optional`, defaults to `'euclidean'`

        :param linkage:
            --------
            |  Which linkage criterion to use. Can be one of `[‘ward’, ‘complete’, ‘average’, ‘single’]`
            |  The linkage criterion determines which distance to use between sets of observation.
            |  The algorithm will merge the pairs of cluster that minimize this criterion.
            |     - `'ward'` minimizes the variance of the clusters being merged.
            |     - `'average`' uses the average of the distances of each observation of the two sets.
            |     - `'complete`' or `'maximum`' linkage uses the maximum distances between all observations of the two \
                sets.
            |     - `'single`' uses the minimum of the distances between all observations of the two sets.
        :type linkage: `str`, `optional`, defaults to `'ward'`

        :param distance_threshold:
            --------
            The linkage distance above which clusters will not be merged.
            If not `None`, `n_clusters` must be `None`.
        :type distance_threshold: `float`, `optional`, defaults to `None`

        :param scaling:
            --------
            Whether to scale the data or not before clustering.

            Formula is `scaled_x = ( x - mean )/standard_deviation`
        :type scaling: `bool`, `optional`, default to `False`

        :return: a CategorizedList that knows the data and their labels
        :rtype: CategorizedList

        """
        skl_cluster = cluster.AgglomerativeClustering(
            n_clusters=n_clusters, affinity=affinity, distance_threshold=distance_threshold, linkage=linkage)
        skl_cluster = cls.fit_cluster(skl_cluster, data.matrix, scaling)
        return cls(data.dessia_objects, skl_cluster.labels_.tolist(), name=name)

    @classmethod
    def from_kmeans(cls, data: HeterogeneousList, n_clusters: int = 2, n_init: int = 10, tol: float = 1e-4,
                    scaling: bool = False, name: str = ""):
        """
        The KMeans algorithm clusters data by trying to separate samples in n groups of equal variance,
        minimizing a criterion known as the inertia or within-cluster sum-of-squares (see below).
        This algorithm requires the number of clusters to be specified. It scales well to large number
        of samples and has been used across a large range of application areas in many different fields.
        The k-means algorithm divides a set of samples into disjoint clusters , each described by the mean
        of the samples in the cluster. The means are commonly called the cluster “centroids”; note that
        they are not, in general, points from, although they live in the same space.
        The K-means algorithm aims to choose centroids that minimise the inertia, or within-cluster
        sum-of-squares criterion.

        See more : https://scikit-learn.org/stable/modules/clustering.html#k-means

        :param data: The future clustered data.
        :type data: List[DessiaObject]

        :param n_clusters:
            --------
            Number of wished clusters
        :type n_clusters: `int`, `optional`, defaults to `2`

        :param n_init:
            --------
            Number of time the k-means algorithm will be run with different centroid seeds.
            The final results will be the best output of n_init consecutive runs in terms of inertia.
        :type n_init: `int`, `optional`, defaults to `10`

        :param tol:
            --------
            Relative tolerance with regards to Frobenius norm of the difference in the cluster centers of two \
                consecutive iterations to declare convergence.
        :type tol: `float`, `optional`, defaults to `1e-4`

        :param scaling:
            --------
            Whether to scale the data or not before clustering.

            Formula is `scaled_x = ( x - mean )/standard_deviation`
        :type scaling: `bool`, `optional`, default to `False`

        :return: a CategorizedList that knows the data and their labels
        :rtype: CategorizedList

        """
        skl_cluster = cluster.KMeans(n_clusters=n_clusters, n_init=n_init, tol=tol)
        skl_cluster = cls.fit_cluster(skl_cluster, data.matrix, scaling)
        return cls(data.dessia_objects, skl_cluster.labels_.tolist(), name=name)

    @classmethod
    def from_dbscan(cls, data: HeterogeneousList, eps: float = 0.5, min_samples: int = 5, mink_power: float = 2,
                    leaf_size: int = 30, metric: str = "euclidean", scaling: bool = False, name: str = ""):
        """
        The DBSCAN algorithm views clusters as areas of high density separated by areas of low density.
        Due to this rather generic view, clusters found by DBSCAN can be any shape, as opposed to k-means
        which assumes that clusters are convex shaped. The central component to the DBSCAN is the concept
        of core samples, which are samples that are in areas of high density. A cluster is therefore a set
        of core samples, each close to each other (measured by some distance measure) and a set of non-core
        samples that are close to a core sample (but are not themselves core samples).
        There are two parameters to the algorithm, min_samples and eps, which define formally what we mean
        when we say dense. Higher min_samples or lower eps indicate higher density necessary to form a cluster.

        See more : https://scikit-learn.org/stable/modules/clustering.html#dbscan

        :param data: The future clustered data.
        :type data: List[DessiaObject]

        :param eps:
            --------
            The maximum distance between two samples for one to be considered as in the neighborhood \
            of the other. This is not a maximum bound on the distances of points within a cluster. This is the most \
            important DBSCAN parameter to choose appropriately for your data set and distance function
        :type eps: `float`, `optional`, defaults to `0.5`

        :param min_samples:
            --------
            The number of samples (or total weight) in a neighborhood for a point to be considered as \
            a core point. This includes the point itself
        :type min_samples: `int`, `optional`, defaults to 5

        :param mink_power:
            --------
            The power of the Minkowski metric to be used to calculate distance between points. \
            If `None`, then `mink_power=2` (equivalent to the Euclidean distance)
        :type mink_power: `float`, `optional`, defaults to `2`

        :param leaf_size:
            --------
            Leaf size passed to BallTree or cKDTree. This can affect the speed of the construction and \
            query, as well as the memory required to store the tree. The optimal value depends on the nature of the \
            problem
        :type leaf_size: `int`, `optional`, defaults to `30`

        :param metric:
            --------
            The metric to use when calculating distance between instances in a feature array. If metric is \
            a string or callable, it must be one of the options allowed by sklearn.metrics.pairwise_distances for its \
            metric parameter. If metric is `'precomputed'`, X is assumed to be a distance matrix and must be square. \
            X may be a sparse graph, in which case only `'nonzero'` elements may be considered neighbors for DBSCAN.
        :type metric: `str`, or `callable`, default to `’euclidean’`


        :param scaling:
            --------
            Whether to scale the data or not before clustering.

            Formula is `scaled_x = ( x - mean )/standard_deviation`
        :type scaling: `bool`, `optional`, default to `False`

        :return: a CategorizedList that knows the data and their labels
        :rtype: CategorizedList

        """
        skl_cluster = cluster.DBSCAN(eps=eps, min_samples=min_samples, p=mink_power, leaf_size=leaf_size, metric=metric)
        skl_cluster = cls.fit_cluster(skl_cluster, data.matrix, scaling)
        return cls(data.dessia_objects, skl_cluster.labels_.tolist(), name=name)

    @classmethod
    def from_pareto_sheets(cls, h_list: HeterogeneousList, costs: List[List[float]], nb_sheets: int = 1):
        """
        Get successive pareto sheets (i.e. optimal points in a DOE for pre-computed costs) and put them in a
        `CategorizedList` where each label is the index of a pareto sheet.

        :param h_list:
            --------
            The HeterogeneousList in which to pick optimal points.
        :type h_list: `HeterogeneousList`

        :param costs:
            --------
            Pre-computed costs of `len(self)`. Can be multi-dimensional.
        :type costs: `List[List[float]]`, `n_samples x n_costs` or `n_costs x n_samples`

        :param nb_sheets:
            --------
            Number of pareto sheets to pick
        :type nb_sheets: `int`, `optional`, default to `1`

        :return: a CategorizedList where each element is labelled with its pareto_sheet. Elements outside a \
        pareto_sheet are labelled `n_sheets`
        :rtype: CategorizedList

        """
        labels = []
        dessia_objects = []
        pareto_sheets, non_optimal_points = h_list.pareto_sheets(costs, nb_sheets)
        for label, pareto_sheet in enumerate(pareto_sheets):
            labels.extend([label] * len(pareto_sheet))
            dessia_objects.extend(pareto_sheet)
        dessia_objects.extend(non_optimal_points)
        labels.extend([len(pareto_sheets)] * len(non_optimal_points))
        return cls(dessia_objects, labels)

    @staticmethod
    def fit_cluster(skl_cluster: cluster, matrix: List[List[float]], scaling: bool):
        if scaling:
            scaled_matrix = HeterogeneousList.scale_data(matrix)
        else:
            scaled_matrix = matrix
        skl_cluster.fit(scaled_matrix)
        return skl_cluster

# Function to implement, to find a good eps parameter for dbscan
# def nearestneighbors(self):
#     vectors = []
#     for machine in self.machines:
#         vector = machine.to_vector()
#         vectors.append(vector)
#     neigh = NearestNeighbors(n_neighbors=14)
#     vectors = StandardScaler().fit_transform(vectors)
#     nbrs = neigh.fit(vectors)
#     distances, indices = nbrs.kneighbors(vectors)
#     distances = npy.sort(distances, axis=0)
#     distances = distances[:, 1]
#     plt.plot(distances)
#     plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
dessia_common

"""

import time
import sys
import warnings
import operator
import math
import random
import itertools

from functools import reduce
import collections
import collections.abc
from copy import deepcopy, copy
import inspect
import json

from typing import List, Dict, Any, Tuple, get_type_hints
import traceback as tb

from importlib import import_module
from ast import literal_eval

import dessia_common.errors
from dessia_common.utils.diff import data_eq, diff, dict_hash, list_hash
from dessia_common.utils.serialization import dict_to_object, serialize_dict_with_pointers, serialize_dict,\
    deserialize_argument, serialize
from dessia_common.utils.types import full_classname, is_sequence, is_bson_valid, TYPES_FROM_STRING
from dessia_common.utils.copy import deepcopy_value
from dessia_common.utils.jsonschema import default_dict, jsonschema_from_annotation, JSONSCHEMA_HEADER,\
    set_default_value
from dessia_common.utils.docstrings import parse_docstring, FAILED_DOCSTRING_PARSING
from dessia_common.exports import XLSXWriter
from dessia_common.typings import JsonSerializable
from dessia_common import templates
from dessia_common.displays import DisplayObject, DisplaySetting
from dessia_common.breakdown import attrmethod_getter, get_in_object_from_path

_FORBIDDEN_ARGNAMES = ['self', 'cls', 'progress_callback', 'return']


def deprecated(use_instead=None):
    def decorated(function):
        def wrapper(*args, **kwargs):
            deprecation_warning(function.__name__, 'Function', use_instead)
            print('Traceback : ')
            tb.print_stack(limit=2)
            return function(*args, **kwargs)

        return wrapper

    return decorated


def deprecation_warning(name, object_type, use_instead=None):
    warnings.simplefilter('once', DeprecationWarning)
    msg = f"\n\n{object_type} {name} is deprecated.\n"
    msg += "It will be removed in a future version.\n"
    if use_instead is not None:
        msg += f"Use {use_instead} instead.\n"
    warnings.warn(msg, DeprecationWarning)
    return msg


class DessiaObject:
    """
    Base class for Dessia's platform compatible objects.
    Gathers generic methods and attributes

    :cvar bool _standalone_in_db:
        Indicates wether class objects should be independant in database or not.
        If False, object will only exist inside its parent.
    :cvar bool _eq_is_data_eq:
        Indicates which type of equality check is used: strict equality or equality based on data.
        If False, Python's object __eq__ method is used (ie. strict), else, user custom data_eq is used (ie. data)
    :cvar List[str] _non_serializable_attributes:
        [Advanced] List of instance attributes that should not be part of serialization with to_dict method.
        These will not be displayed in platform object tree, for instance.
    :cvar List[str] _non_data_eq_attributes:
        [Advanced] List of instance attributes that should not be part of equality check with data__eq__ method
        (if _eq_is_data_eq is True).
    :cvar List[str] _non_data_hash_attributes:
        [Advanced] List of instance attributes that should not be part of hash computation with data__hash__ method
        (if _eq_is_data_eq is True).
    :cvar List[str] _ordered_attributes:
        Documentation not available yet.
    :cvar List[str] _titled_attributes:
        Documentation not available yet.
    :cvar List[str] _init_variables:
        Documentation not available yet.
    :cvar List[str] _export_formats:
        List of all available export formats. Class must define a export_[format] for each format in _export_formats
    :cvar List[str] _allowed_methods:
        List of all methods that are runnable from platform.
    :cvar List[str] _whitelist_attributes:
        Documentation not available yet.
    :cvar List[str] _whitelist_attributes: List[str]


    :ivar str name: Name of object.
    :ivar Any kwargs: Additionnal user metadata
    """
    _standalone_in_db = False
    _non_serializable_attributes = []
    _non_editable_attributes = []
    _non_data_eq_attributes = ['name']
    _non_data_hash_attributes = ['name']
    _ordered_attributes = []
    _titled_attributes = []
    _eq_is_data_eq = True
    _vector_features = None

    _init_variables = None
    _allowed_methods = []
    _whitelist_attributes = []

    def __init__(self, name: str = '', **kwargs):
        """
        Generic init of DessiA Object. Only store name in self. To be overload and call in specific class init
        """
        self.name = name
        for property_name, property_value in kwargs.items():
            setattr(self, property_name, property_value)

    def __hash__(self):
        """
        Compute a int from object
        """
        if self._eq_is_data_eq:
            return self._data_hash()
        return object.__hash__(self)

    def __eq__(self, other_object):
        """
        Generic equality of two objects. behavior can be controled by class attribute _eq_is_data_eq to tell
        if we must use python equality (based on memory addresses) (_eq_is_data_eq = False) or a data equality (True)
        """
        if self._eq_is_data_eq:
            if self.__class__.__name__ != other_object.__class__.__name__:
                return False
            if self._data_hash() != other_object._data_hash():
                return False
            return self._data_eq(other_object)
        return object.__eq__(self, other_object)

    def _data_eq(self, other_object):
        return data_eq(self, other_object)

    def _data_hash(self):
        hash_ = 0
        forbidden_keys = (self._non_data_eq_attributes + self._non_data_hash_attributes + ['package_version', 'name'])
        for key, value in self._serializable_dict().items():
            if key not in forbidden_keys:
                if is_sequence(value):
                    hash_ += list_hash(value)
                elif isinstance(value, dict):
                    hash_ += dict_hash(value)
                elif isinstance(value, str):
                    hash_ += sum(ord(v) for v in value)
                else:
                    hash_ += hash(value)
        return int(hash_ % 1e5)

    def _data_diff(self, other_object):
        """
        Make a diff between two objects
        returns: different values, missing keys in other object
        """
        # return diff(self, other_object)
        return diff(self, other_object)

    def _get_from_path(self, path: str):
        return get_in_object_from_path(self, path)

    @property
    def full_classname(self):
        """
        Full classname of class like: package.module.submodule.classname
        """
        return full_classname(self)

    def base_dict(self):
        """
        A base dict for to_dict: put name, object class and version in a dict
        """
        package_name = self.__module__.split('.', maxsplit=1)[0]
        if package_name in sys.modules:
            package = sys.modules[package_name]
            if hasattr(package, '__version__'):
                package_version = package.__version__
            else:
                package_version = None
        else:
            package_version = None

        object_class = self.__module__ + '.' + self.__class__.__name__
        dict_ = {'name': self.name, 'object_class': object_class}
        if package_version:
            dict_['package_version'] = package_version
        return dict_

    def _serializable_dict(self):
        """
        Returns a dict of attribute_name, values (still python, not serialized)
        Keys are filtered with non serializable attributes controls
        """

        dict_ = {k: v for k, v in self.__dict__.items()
                 if k not in self._non_serializable_attributes and not k.startswith('_')}
        return dict_

    def to_dict(self, use_pointers: bool = True, memo=None, path: str = '#') -> JsonSerializable:
        """
        Generic to_dict method
        """
        if memo is None:
            memo = {}

        # Default to dict
        serialized_dict = self.base_dict()
        dict_ = self._serializable_dict()
        if use_pointers:
            serialized_dict.update(serialize_dict_with_pointers(dict_, memo, path)[0])
        else:
            serialized_dict.update(serialize_dict(dict_))

        return serialized_dict

    @classmethod
    def dict_to_object(cls, dict_: JsonSerializable, force_generic: bool = False, global_dict=None,
                       pointers_memo: Dict[str, Any] = None, path: str = '#') -> 'DessiaObject':
        """
        Generic dict_to_object method
        """
        if cls is not DessiaObject:
            obj = dict_to_object(dict_=dict_, class_=cls, force_generic=force_generic, global_dict=global_dict,
                                 pointers_memo=pointers_memo, path=path)
            return obj

        if 'object_class' in dict_:
            obj = dict_to_object(dict_=dict_, force_generic=force_generic, global_dict=global_dict,
                                 pointers_memo=pointers_memo, path=path)
            return obj

        raise NotImplementedError('No object_class in dict')

    @classmethod
    def base_jsonschema(cls):
        jsonschema = deepcopy(JSONSCHEMA_HEADER)
        jsonschema['properties']['name'] = {
            "type": 'string',
            "title": "Object Name",
            "description": "Object name",
            "editable": True,
            "default_value": "Object Name"
        }
        return jsonschema

    @classmethod
    def jsonschema(cls):
        """
        Jsonschema of class: transfer python data structure to web standard
        """
        if hasattr(cls, '_jsonschema'):
            _jsonschema = cls._jsonschema
            return _jsonschema

        # Get __init__ method and its annotations
        init = cls.__init__
        if cls._init_variables is None:
            annotations = get_type_hints(init)
        else:
            annotations = cls._init_variables

        # Get ordered variables
        if cls._ordered_attributes:
            ordered_attributes = cls._ordered_attributes
        else:
            ordered_attributes = list(annotations.keys())

        unordered_count = 0

        # Parse docstring
        try:
            docstring = cls.__doc__
            parsed_docstring = parse_docstring(docstring=docstring, annotations=annotations)
        except Exception:
            parsed_docstring = FAILED_DOCSTRING_PARSING
        parsed_attributes = parsed_docstring['attributes']

        # Initialize jsonschema
        _jsonschema = deepcopy(JSONSCHEMA_HEADER)

        required_arguments, default_arguments = inspect_arguments(method=init, merge=False)
        _jsonschema['required'] = required_arguments
        _jsonschema['standalone_in_db'] = cls._standalone_in_db
        _jsonschema['description'] = parsed_docstring['description']
        _jsonschema['python_typing'] = str(cls)

        # Set jsonschema
        for annotation in annotations.items():
            name = annotation[0]
            if name in ordered_attributes:
                order = ordered_attributes.index(name)
            else:
                order = len(ordered_attributes) + unordered_count
                unordered_count += 1
            if name in cls._titled_attributes:
                title = cls._titled_attributes[name]
            else:
                title = None

            if name != 'return':
                editable = name not in cls._non_editable_attributes
                annotation_type = type_from_annotation(annotation[1], cls)
                annotation = (name, annotation_type)
                jss_elt = jsonschema_from_annotation(annotation=annotation, jsonschema_element={}, order=order,
                                                     editable=editable, title=title,
                                                     parsed_attributes=parsed_attributes)
                _jsonschema['properties'].update(jss_elt)
                if name in default_arguments:
                    default = set_default_value(_jsonschema['properties'], name, default_arguments[name])
                    _jsonschema['properties'].update(default)

        _jsonschema['classes'] = [cls.__module__ + '.' + cls.__name__]
        _jsonschema['whitelist_attributes'] = cls._whitelist_attributes
        return _jsonschema

    @property
    def _method_jsonschemas(self):
        """
        Generates dynamic jsonschemas for methods of class
        """
        jsonschemas = {}
        class_ = self.__class__

        # TOCHECK Backward compatibility. Will need to be changed
        if hasattr(class_, '_dessia_methods'):
            allowed_methods = class_._dessia_methods
        else:
            allowed_methods = class_._allowed_methods

        valid_method_names = [m for m in dir(class_) if not m.startswith('_') and m in allowed_methods]

        for method_name in valid_method_names:
            method = getattr(class_, method_name)

            if not isinstance(method, property):
                required_args, default_args = inspect_arguments(method=method, merge=False)
                annotations = get_type_hints(method)
                if annotations:
                    jsonschemas[method_name] = deepcopy(JSONSCHEMA_HEADER)
                    jsonschemas[method_name]['required'] = []
                    jsonschemas[method_name]['is_method'] = True
                    for i, annotation in enumerate(annotations.items()):
                        # TOCHECK Not actually ordered
                        argname = annotation[0]
                        if argname not in _FORBIDDEN_ARGNAMES:
                            if argname in required_args:
                                jsonschemas[method_name]['required'].append(str(i))
                            jsonschema_element = jsonschema_from_annotation(annotation, {}, i)[argname]

                            jsonschemas[method_name]['properties'][str(i)] = jsonschema_element
                            if argname in default_args:
                                default = set_default_value(jsonschemas[method_name]['properties'],
                                                            str(i),
                                                            default_args[argname])
                                jsonschemas[method_name]['properties'].update(default)
        return jsonschemas

    def method_dict(self, method_name=None, method_jsonschema=None):
        """
        Return a jsonschema of a method arguments
        """
        if method_name is None and method_jsonschema is None:
            msg = 'No method name nor jsonschema provided'
            raise NotImplementedError(msg)

        if method_name is not None and method_jsonschema is None:
            method_jsonschema = self._method_jsonschemas[method_name]

        dict_ = default_dict(method_jsonschema)
        return dict_

    def dict_to_arguments(self, dict_, method):
        """
        Transform serialized argument of a method to python objects ready to use in method evaluation
        """
        method_object = getattr(self, method)
        args_specs = inspect.getfullargspec(method_object)
        allowed_args = args_specs.args[1:]

        arguments = {}
        for i, arg in enumerate(allowed_args):
            if str(i) in dict_:
                arg_specs = args_specs.annotations[arg]
                value = dict_[str(i)]
                try:
                    deserialized_value = deserialize_argument(arg_specs, value)
                except TypeError as err:
                    msg = 'Error in deserialisation of value: '
                    msg += f'{value} of expected type {arg_specs}'
                    raise TypeError(msg) from err
                arguments[arg] = deserialized_value
        return arguments

    def save_to_file(self, filepath: str, indent: int = 2):
        """
        Save object to a JSON file
        :param filepath: either a string reprensenting the filepath or a stream
        """
        if not filepath.endswith('.json'):
            filepath += '.json'
            print(f'Changing name to {filepath}')
        with open(filepath, 'w', encoding='utf-8') as file:
            self.save_to_stream(file, indent=indent)

    def save_to_stream(self, stream, indent: int = 2):
        try:
            dict_ = self.to_dict(use_pointers=True)
        except TypeError:
            dict_ = self.to_dict()

        json.dump(dict_, stream, indent=indent)

    @classmethod
    def load_from_stream(cls, stream):
        dict_ = json.loads(stream.read().decode('utf-8'))
        return cls.dict_to_object(dict_)

    @classmethod
    def load_from_file(cls, filepath: str):
        """
        Load object from a json file
        :param filepath: either a string reprensenting the filepath or a stream
        """
        with open(filepath, 'r', encoding='utf-8') as file:
            dict_ = json.load(file)

        return cls.dict_to_object(dict_)

    def is_valid(self):
        return True

    def copy(self, deep=True, memo=None):
        if deep:
            return deepcopy(self, memo=memo)
        return copy(self)

    def __copy__(self):
        """
        Generic copy use inits of objects
        """
        class_argspec = inspect.getfullargspec(self.__class__)
        dict_ = {}
        for arg in class_argspec.args:
            if arg != 'self':
                value = self.__dict__[arg]
                if hasattr(value, '__copy__'):
                    dict_[arg] = value.__copy__()
                else:
                    dict_[arg] = value
        return self.__class__(**dict_)

    def __deepcopy__(self, memo=None):
        """
        Generic deep copy use inits of objects
        """
        class_argspec = inspect.getfullargspec(self.__class__)
        if memo is None:
            memo = {}
        dict_ = {}
        for arg in class_argspec.args:
            if arg != 'self':
                dict_[arg] = deepcopy_value(getattr(self, arg), memo=memo)
        return self.__class__(**dict_)

    def plot_data(self):  # TODO: Should it have a **kwargs argument ?
        return []

    def plot(self, **kwargs):
        """
        Generic plot getting plot_data function to plot
        """
        if hasattr(self, 'plot_data'):
            import plot_data
            for data in self.plot_data(**kwargs):  # TODO solve inconsistence with the plot_data method just above
                plot_data.plot_canvas(plot_data_object=data,
                                      canvas_id='canvas',
                                      width=1400, height=900,
                                      debug_mode=False)
        else:
            msg = 'Class {} does not implement a plot_data method to define what to plot'
            raise NotImplementedError(msg.format(self.__class__.__name__))

    def mpl_plot(self, **kwargs):
        """
        Plot with matplotlib using plot_data function
        """
        axs = []
        if hasattr(self, 'plot_data'):
            try:
                plot_datas = self.plot_data(**kwargs)
            except TypeError as error:
                raise TypeError(f'{self.__class__.__name__}.{error}') from error
            for data in plot_datas:
                if hasattr(data, 'mpl_plot'):
                    ax = data.mpl_plot()
                    axs.append(ax)
        else:
            msg = 'Class {} does not implement a plot_data method to define what to plot'
            raise NotImplementedError(msg.format(self.__class__.__name__))

        return axs

    @staticmethod
    def display_settings() -> List[DisplaySetting]:
        """
        Returns a list of json describing how to call subdisplays
        """
        return [DisplaySetting('markdown', 'markdown', 'to_markdown', None),
                DisplaySetting('plot_data', 'plot_data', 'plot_data', None, serialize_data=True)]

    def _display_from_selector(self, selector: str, **kwargs) -> DisplayObject:
        """
        Generate the display from the selector
        """
        reference_path = kwargs.get('reference_path', '')

        display_setting = self._display_settings_from_selector(selector)
        track = ''
        try:
            data = attrmethod_getter(self, display_setting.method)(**display_setting.arguments)
        except:
            data = None
            track = tb.format_exc()

        if display_setting.serialize_data:
            data = serialize(data)
        return DisplayObject(type_=display_setting.type, data=data, reference_path=reference_path, traceback=track)

    def _display_settings_from_selector(self, selector: str):
        for display_setting in self.display_settings():
            if display_setting.selector == selector:
                return display_setting
        raise ValueError(f"No such selector '{selector}' in display of class '{self.__class__.__name__}'")

    def _displays(self, **kwargs) -> List[JsonSerializable]:
        """
        Generate displays of the object to be plot in the DessiA Platform
        """
        reference_path = kwargs.get('reference_path', '')

        displays = []
        for display_setting in self.display_settings():
            display_ = self._display_from_selector(display_setting.selector, reference_path=reference_path)
            displays.append(display_.to_dict())
        return displays

    def to_markdown(self) -> str:
        """
        Render a markdown of the object output type: string
        """
        return templates.dessia_object_markdown_template.substitute(name=self.name, class_=self.__class__.__name__)

    def _performance_analysis(self):
        """
        Prints time of rendering some commons operations (serialization, hash, displays)
        """
        data_hash_time = time.time()
        self._data_hash()
        data_hash_time = time.time() - data_hash_time
        print(f'Data hash time: {round(data_hash_time, 3)} seconds')

        todict_time = time.time()
        dict_ = self.to_dict()
        todict_time = time.time() - todict_time
        print(f'to_dict time: {round(todict_time, 3)} seconds')

        dto_time = time.time()
        self.dict_to_object(dict_)
        dto_time = time.time() - dto_time
        print(f'dict_to_object time: {round(dto_time, 3)} seconds')

        for display_setting in self.display_settings():
            display_time = time.time()
            self._display_from_selector(display_setting.selector)
            display_time = time.time() - display_time
            print(f'Generation of display {display_setting.selector} in: {round(display_time, 6)} seconds')

    def _check_platform(self):
        """
        Reproduce lifecycle on platform (serialization, display)
        raise an error if something is wrong
        """
        try:
            dict_ = self.to_dict(use_pointers=True)
        except TypeError:
            dict_ = self.to_dict()
        json_dict = json.dumps(dict_)
        decoded_json = json.loads(json_dict)
        deserialized_object = self.dict_to_object(decoded_json)
        if not deserialized_object._data_eq(self):
            print('data diff: ', self._data_diff(deserialized_object))
            raise dessia_common.errors.DeserializationError('Object is not equal to itself'
                                                            ' after serialization/deserialization')
        copied_object = self.copy()
        if not copied_object._data_eq(self):
            try:
                print('data diff: ', self._data_diff(copied_object))
            except:
                pass
            raise dessia_common.errors.CopyError('Object is not equal to itself after copy')

        valid, hint = is_bson_valid(stringify_dict_keys(dict_))
        if not valid:
            raise ValueError(hint)
        json.dumps(self._displays())
        json.dumps(self._method_jsonschemas)

    def to_xlsx(self, filepath: str):
        """
        Exports the object to an XLSX file given by the filepath
        """
        with open(filepath, 'wb') as file:
            self.to_xlsx_stream(file)

    def to_xlsx_stream(self, stream):
        """
        Exports the object to an XLSX to a given stream
        """
        writer = XLSXWriter(self)
        writer.save_to_stream(stream)

    def _export_formats(self):
        formats = [{"extension": "json", "method_name": "save_to_stream", "text": True, "args": {}},
                   {"extension": "xlsx", "method_name": "to_xlsx_stream", "text": False, "args": {}}]
        return formats

    def to_vector(self):
        vectored_objects = []
        for feature in self.vector_features():
            vectored_objects.append(getattr(self, feature.lower()))
            if not hasattr(self, feature.lower()):
                raise NotImplementedError(f"{feature} is not an attribute for {self.__class__.__name__} objects. " +
                                          f"<to_vector> method must be customized in {self.__class__.__name__} to " +
                                          "handle computed values that are not class or instance attributes.")
        return vectored_objects

    @classmethod
    def vector_features(cls):
        if cls._vector_features is None:
            return list(set(get_attribute_names(cls)).difference(get_attribute_names(DessiaObject)))
        return cls._vector_features


class PhysicalObject(DessiaObject):
    """
    Represent an object with CAD capabilities
    """

    @staticmethod
    def display_settings():
        """
        Returns a list of json describing how to call subdisplays
        """
        display_settings = DessiaObject.display_settings()
        display_settings.append(DisplaySetting(selector='cad', type_='babylon_data',
                                               method='volmdlr_volume_model().babylon_data', serialize_data=True))
        return display_settings

    def volmdlr_primitives(self):
        """
        Return a list of volmdlr primitives to build up volume model
        """
        return []

    def volmdlr_volume_model(self, **kwargs):
        """
        Gives the volmdlr VolumeModel
        """
        import volmdlr as vm  # !!! Avoid circular imports, is this OK ?
        return vm.core.VolumeModel(self.volmdlr_primitives(**kwargs))

    def to_step(self, filepath: str):
        """
        Exports the CAD of the object to step. Works if the class define a custom volmdlr model
        :param filepath: a str representing a filepath
        """
        self.volmdlr_volume_model().to_step(filepath=filepath)
        return self.volmdlr_volume_model().to_step(filepath=filepath)

    def to_step_stream(self, stream):
        """
        Exports the CAD of the object to a stream in the STEP format. Works if the class define a custom volmdlr model
        """
        return self.volmdlr_volume_model().to_step_stream(stream=stream)

    def to_stl_stream(self, stream):
        """
        Exports the CAD of the object to STL to a given stream
        """
        return self.volmdlr_volume_model().to_stl_stream(stream=stream)

    def to_stl(self, filepath):
        """
        Exports the CAD of the object to STL. Works if the class define a custom volmdlr model
        :param filepath: a str representing a filepath
        """
        return self.volmdlr_volume_model().to_stl(filepath=filepath)

    # def _displays(self, **kwargs):
    #     """
    #     Compute the list of displays
    #     """
    #     return DessiaObject._displays(self, **kwargs)

    def babylonjs(self, use_cdn=True, debug=False, **kwargs):
        """
        Show the 3D volmdlr of an object by calling volmdlr_volume_model method
        and plot in in browser
        """
        self.volmdlr_volume_model(**kwargs).babylonjs(use_cdn=use_cdn, debug=debug)

    def save_babylonjs_to_file(self, filename: str = None, use_cdn: bool = True, debug: bool = False, **kwargs):
        self.volmdlr_volume_model(**kwargs).save_babylonjs_to_file(filename=filename, use_cdn=use_cdn, debug=debug)

    def _export_formats(self):
        formats = DessiaObject._export_formats(self)
        formats3d = [{"extension": "step", "method_name": "to_step_stream", "text": True, "args": {}},
                     {"extension": "stl", "method_name": "to_stl_stream", "text": False, "args": {}}]
        formats.extend(formats3d)
        return formats


class MovingObject(PhysicalObject):

    def volmdlr_primitives_step_frames(self):
        """
        Return a list of volmdlr primitives to build up volume model
        """
        raise NotImplementedError('Object inheriting MovingObject should implement volmdlr_primitives_step_frames')

    def volmdlr_volume_model(self, **kwargs):
        import volmdlr as vm  # !!! Avoid circular imports, is this OK ?
        return vm.core.MovingVolumeModel(self.volmdlr_primitives(**kwargs),
                                         self.volmdlr_primitives_step_frames(**kwargs))


class Parameter(DessiaObject):
    def __init__(self, lower_bound, upper_bound, periodicity=None, name=''):
        DessiaObject.__init__(self, name=name)
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.periodicity = periodicity

    def random_value(self):
        """
        Sample a value within the bounds
        """
        return random.uniform(self.lower_bound, self.upper_bound)

    def are_values_equal(self, value1, value2, tol=1e-2):
        if self.periodicity is not None:
            value1 = value1 % self.periodicity
            value2 = value2 % self.periodicity

        return math.isclose(value1, value2, abs_tol=tol)

    def normalize(self, value):
        normalized_value = (value - self.lower_bound) / (self.upper_bound - self.lower_bound)
        return normalized_value

    def original_value(self, normalized_value):
        value = normalized_value * (self.upper_bound - self.lower_bound) + self.lower_bound
        return value

    def optimizer_bounds(self):
        if self.periodicity is not None:
            return (self.lower_bound - 0.5 * self.periodicity, self.upper_bound + 0.5 * self.periodicity)
        return None


class ParameterSet(DessiaObject):
    def __init__(self, values, name=''):
        self.values = values

        DessiaObject.__init__(self, name=name)

    @property
    def parameters(self):
        parameters = [Parameter(min(v), max(v), name=k) for k, v in self.values.items()]
        return parameters

    @property
    def means(self):
        means = {k: sum(v) / len(v) for k, v in self.values.items()}
        return means


class DessiaFilter(DessiaObject):
    """
    Base class for filters working on lists of DessiaObjects (List[DessiaObject]).

    :param attribute:
        --------
        Name of attribute on which to filter
    :type attribute: str

    :param comparison_operator:
        --------
        Comparison operator
    :type comparison_operator: str

    :param bound:
        --------
        The bound value to compare `'attribute'` of DessiaObjects of a list with `'comparison_operator'`
    :type bound: float

    :param name:
        --------
        Name of filter
    :type name: `str`, `optional`, defaults to `''`

    :Comparison operators:
        * greater than: >=, gte, ge
        * greater: >, gt
        * lower than: <=, lte, le
        * lower: <, lt
        * equal: ==, eq
        * different: !=, ne
    """

    _REAL_OPERATORS = {'>': operator.gt, '<': operator.lt, '>=': operator.ge, '<=': operator.le, '==': operator.eq,
                       '!=': operator.ne, 'gt': operator.gt, 'lt': operator.lt, 'ge': operator.ge, 'le': operator.le,
                       'eq': operator.eq, 'ne': operator.ne, 'gte': operator.ge, 'lte': operator.le}

    def __init__(self, attribute: str, comparison_operator: str, bound: float, name: str = ''):
        self.attribute = attribute
        self.comparison_operator = comparison_operator
        self.bound = bound
        DessiaObject.__init__(self, name=name)

    def __str__(self, offset_attr: int = 10, offset_boun: int = 0):
        offset_oper = 0
        if offset_boun == 0:
            offset_boun = len(str(self.bound)) + 2
        string_operator = {'>': '>', '<': '<', '>=': '>=', '<=': '<=', '==': '==', '!=': '!=', 'gt': '>', 'lt': '<',
                           'ge': '>=', 'le': '<=', 'eq': '==', 'ne': '!=', 'gte': '>=', 'lte': '<='}
        printed_operator = string_operator[self.comparison_operator]
        return (self.attribute + " " * (offset_attr - len(self.attribute)) +
                printed_operator + " " * (offset_oper - len(printed_operator)) +
                " " * (offset_boun - len(str(self.bound))) + str(self.bound))

    def __hash__(self):
        hash_ = len(self.attribute)
        hash_ += hash(self.comparison_operator)
        hash_ += hash(self.bound)
        return int(hash_)

    def __eq__(self, other: 'DessiaFilter'):
        same_attr = self.attribute == other.attribute
        same_op = self.comparison_operator == other.comparison_operator
        same_bound = self.bound == other.bound
        return same_attr and same_op and same_bound

    def _comparison_operator(self):
        return self._REAL_OPERATORS[self.comparison_operator]

    def _to_lambda(self):
        return lambda x: (self._comparison_operator()(enhanced_deep_attr(value, self.attribute), self.bound)
                          for value in x)

    def get_booleans_index(self, values: List[DessiaObject]):
        """
        Get the boolean indexing of a filtered list

        :param values:
            List of DessiaObjects to filter
        :type values: List[DessiaObject]

        :return: `list of length `len(values)` where elements are `True` if kept by the filter, otherwise `False`.
        :rtype: `List[bool]`

        Examples
        --------
        >>> from dessia_common.core import DessiaFilter
        >>> from dessia_common.models import all_cars_no_feat
        >>> values = all_cars_no_feat[:5]
        >>> filter_ = DessiaFilter('weight', '<=', 3500.)
        >>> filter_.get_booleans_index(values)
        [False, False, True, True, True]
        """
        return list(self._to_lambda()(values))

    @staticmethod
    def booleanlist_to_indexlist(booleans_list: List[int]):  # TODO: Should it exist ?
        """
        Transform a boolean list to an index list

        :param booleans_list:
            list of length `len(values)` where elements are `True` if kept, otherwise `False`.
        :type booleans_list: List[int]

        :return: list of kept indexes
        :rtype: List[int]

        Examples
        --------
        >>> from dessia_common.core import DessiaFilter
        >>> from dessia_common.models import all_cars_no_feat
        >>> values = all_cars_no_feat[:5]
        >>> filter_ = DessiaFilter('weight', '<=', 3500.)
        >>> booleans_list = filter_.get_booleans_index(values)
        [False, False, True, True, True]
        >>> DessiaFilter.booleanlist_to_indexlist(booleans_list)
        [2, 3, 4]
        """
        return list(itertools.compress(range(len(booleans_list)), booleans_list))

    @staticmethod
    def apply(values: List[DessiaObject], booleans_list: List[List[bool]]):
        """
        Apply a Dessia Filter on a list of DessiaObjects

        :param values:
            --------
            List of DessiaObjects to filter
        :type values: List[DessiaObject]

        :param booleans_list:
            --------
            list of length `len(values)` where elements are `True` if kept, otherwise `False`.
        :type booleans_list: List[List[bool]]

        :return: List of filtered values
        :rtype: List[DessiaObject]

        Examples
        --------
        >>> from dessia_common.core import DessiaFilter
        >>> from dessia_common.models import all_cars_no_feat
        >>> values = all_cars_no_feat[:5]
        >>> filter_ = DessiaFilter('weight', '<=', 3500.)
        >>> booleans_list = filter_.get_booleans_index(values)
        [False, False, True, True, True]
        >>> for car in DessiaFilter.apply(values, booleans_list): print(car.weight)
        3436.0, 3433.0, 3449.0
        """
        return list(itertools.compress(values, booleans_list))


class FiltersList(DessiaObject):
    """
    Combine several filters stored as a list of DessiaFilters with a logical operator.

    :param filters:
        --------
        List of DessiaFilters to combine
    :type filters: List[DessiaFilter]

    :param logical_operator:
        --------
        Logical operator to combine filters
    :type logical_operator: str

    :param name:
        ---------
        Name of FiltersList
    :type name: `str`, `optional`, defaults to `''`

    :Logical operators: `'and'`, `'or'`, `'xor'`
    """
    _standalone_in_db = True

    def __init__(self, filters: List[DessiaFilter], logical_operator: str = 'and', name: str = ''):
        self.filters = filters
        self.logical_operator = logical_operator
        DessiaObject.__init__(self, name=name)

    def __len__(self):
        return len(self.filters)

    def __str__(self):
        print_lim = 15
        len_attr = max(map(len, [filter_.attribute for filter_ in self.filters]))
        len_numb = max(map(len, [str(filter_.bound) for filter_ in self.filters]))
        prefix = f"{self.__class__.__name__} {self.name if self.name != '' else hex(id(self))}: "
        prefix += f"{len(self)} filters combined with '" + self.logical_operator + "' operator :\n"
        string = ""
        for filter_ in self.filters[:print_lim]:
            string += " " * 3 + "- "
            string += filter_.__str__(len_attr + 2, len_numb + 2)
            string += "\n"
        return prefix + string

    @classmethod
    def from_filters_list(cls, filters: List[DessiaFilter], logical_operator: str = 'and', name: str = ''):
        """
        Compute a FilersList from a pre-built list of DessiaFilter

        :param filters:
            --------
            List of DessiaFilters to combine
        :type filters: List[DessiaFilter]

        :param logical_operator:
            --------
            Logical operator to combine filters (`'and'`, `'or'` or `'xor'`)
        :type logical_operator: `str`, `optional`, defaults to `'and'`

        :param name:
            --------
            Name of FiltersList
        :type name: `str`, `optional`, defaults to `''`

        :return: A new instantiated list of DessiaFilter
        :rtype: FiltersList

        Examples
        --------
        >>> from dessia_common.core import DessiaFilter, FiltersList
        >>> filters = [DessiaFilter('weight', '<=', 3500.), DessiaFilter('mpg', '<=', 40.)]
        >>> filters_list = FiltersList(filters, logical_operator="or", name="example")
        >>> print(filters_list)
        FiltersList example: 2 filters combined with 'or' operator :
           - weight  <=  3500.0
           - mpg     <=    40.0
        """
        return cls(filters=filters, logical_operator=logical_operator, name=name)

    @staticmethod
    def combine_booleans_lists(booleans_lists: List[List[bool]], logical_operator: str = "and"):
        """
        Combine a list of `n` booleans indexes with the logical operator into a simple booleans index.

        :param booleans_lists:
            --------
            List of `n` booleans indexes
        :type booleans_lists: List[List[bool]]

        :param logical_operator:
            --------
            Logical operator to combine filters (`'or'`, `'and'` or `'xor'`)
        :type logical_operator: `str`, `optional`, defaults to 'and'

        :raises NotImplementedError: If logical_operator is not one of `'and'`, `'or'`, `'xor'`, raises an error

        :return: Booleans index of the filtered data
        :rtype: List[bool]

        Examples
        --------
        >>> from dessia_common.core import FiltersList
        >>> booleans_lists = [[True, True, False, False], [False, True, True, False]]
        >>> FiltersList.combine_booleans_lists(booleans_lists, logical_operator="xor")
        [True, False, True, False]
        """
        if logical_operator == 'and':
            return [all(booleans_tuple) for booleans_tuple in zip(*booleans_lists)]
        if logical_operator == 'or':
            return [any(booleans_tuple) for booleans_tuple in zip(*booleans_lists)]
        if logical_operator == 'xor':
            return [True if sum(booleans_tuple) == 1 else False for booleans_tuple in zip(*booleans_lists)]
        raise NotImplementedError(f"'{logical_operator}' str for 'logical_operator' attribute is not a use case")

    def get_booleans_index(self, dobjects_list: List[DessiaObject]):
        """
        Compute all the filters of `self.filters` on `dobjects_list` and returns a booleans index of `dobjects_list`

        :param dobject_list: List of data to filter
        :type dobject_list: List[DessiaObject]

        :return: A `booleans index` of `dobjects_list` of the list of data to filter (`dobjects_list`)
        :rtype: List[bool]

        Examples
        --------
        >>> from dessia_common.core import FiltersList
        >>> from dessia_common.models import all_cars_no_feat
        >>> dobjects_list = all_cars_no_feat[:5]
        >>> filters = [DessiaFilter('weight', '<=', 4000.), DessiaFilter('mpg', '>=', 30.)]
        >>> filters_list = FiltersList(filters, logical_operator="xor", name="example")
        >>> filters_list.get_booleans_index(dobjects_list)
        [True, True, True, True, True]
        """
        booleans_index = []
        for filter_ in self.filters:
            booleans_index.append(filter_.get_booleans_index(dobjects_list))
        return self.__class__.combine_booleans_lists(booleans_index, self.logical_operator)

    def apply(self, dobjects_list: List[DessiaObject]):
        """
        Apply a FiltersList on a list of DessiaObjects

        :param dobjects_list: List of DessiaObjects to filter
        :type dobjects_list: List[DessiaObject]

        :return: List of filtered values
        :rtype: List[DessiaObject]

        Examples
        --------
        >>> from dessia_common.core import FiltersList
        >>> from dessia_common.datatools import HeterogeneousList
        >>> from dessia_common.models import all_cars_wi_feat
        >>> filters = [DessiaFilter('weight', '<=', 1650.), DessiaFilter('mpg', '>=', 45.)]
        >>> filters_list = FiltersList(filters, logical_operator="xor", name="example")
        >>> filtered_cars = filters_list.apply(all_cars_wi_feat)
        >>> print(HeterogeneousList(filtered_cars, name="example"))
        HeterogeneousList example: 3 samples, 5 features
        |         Mpg         |    Displacement    |     Horsepower     |       Weight       |    Acceleration    |
        -----------------------------------------------------------------------------------------------------------
        |               35.0  |             0.072  |              69.0  |            1613.0  |              18.0  |
        |               31.0  |             0.076  |              52.0  |            1649.0  |              16.5  |
        |               46.6  |             0.086  |              65.0  |            2110.0  |              17.9  |
        """
        booleans_index = self.get_booleans_index(dobjects_list)
        return DessiaFilter.apply(dobjects_list, booleans_index)


def dict_merge(old_dct, merge_dct, add_keys=True, extend_lists=True):
    """ Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.

    This version will return a copy of the dictionary and leave the original
    arguments untouched.

    The optional argument ``add_keys``, determines whether keys which are
    present in ``merge_dct`` but not ``dct`` should be included in the
    new dict.

    Args:
        old_dct (dict) onto which the merge is executed
        merge_dct (dict): dct merged into dct
        add_keys (bool): whether to add new keys
        extend_lists (bool) : wether to extend lists if keys are updated
                              and value is a list

    Returns:
        dict: updated dict
    """
    dct = deepcopy(old_dct)
    if not add_keys:
        merge_dct = {k: merge_dct[k] for k in set(dct).intersection(set(merge_dct))}

    for key, value in merge_dct.items():
        if isinstance(dct.get(key), dict) and isinstance(value, collections.abc.Mapping):
            dct[key] = dict_merge(dct[key], merge_dct[key], add_keys=add_keys, extend_lists=extend_lists)
        elif isinstance(dct.get(key), list) and extend_lists:
            dct[key].extend(value)
        else:
            dct[key] = value

    return dct


def stringify_dict_keys(obj):
    if isinstance(obj, (list, tuple)):
        new_obj = []
        for elt in obj:
            new_obj.append(stringify_dict_keys(elt))

    elif isinstance(obj, dict):
        new_obj = {}
        for key, value in obj.items():
            new_obj[str(key)] = stringify_dict_keys(value)
    else:
        return obj
    return new_obj


def getdeepattr(obj, attr):
    return reduce(getattr, [obj] + attr.split('.'))


def enhanced_deep_attr(obj, sequence):
    """
    Deprecated. Use get_in_from_path from dessia_common.breakdown.py instead

    Get deep attribute where Objects, Dicts and Lists can be found in recursion.

    :param obj: Parent object in which recursively find attribute represented by sequence
    :param sequence: List of strings and integers that represents path to deep attribute.
    :return: Value of deep attribute
    """
    warnings.warn("enhanced_deep_attr is deprecated. Use get_in_from_path from dessia_common.breakdown.py instead")
    if isinstance(sequence, str):
        path = f"#/{sequence}"
    else:
        path = f"#/{'/'.join(sequence)}"
    return get_in_object_from_path(object_=obj, path=path)

        # # Sequence is a string and not a sequence of deep attributes
        # if '/' in sequence:
        #     # Is deep attribute reference
        #     sequence = sequence.split('/')
        #     return enhanced_deep_attr(obj=obj, sequence=sequence)
        # # Is direct attribute
        # return enhanced_get_attr(obj=obj, attr=sequence)
    #
    # # Get direct attribute
    # subobj = enhanced_get_attr(obj=obj, attr=sequence[0])
    # if len(sequence) > 1:
    #     # Recursively get deep attributes
    #     subobj = enhanced_deep_attr(obj=subobj, sequence=sequence[1:])
    # return subobj


def enhanced_get_attr(obj, attr):
    """
    Safely get attribute in obj. Obj can be of Object, Dict, or List type

    :param obj: Parent object in which find given attribute
    :param attr: String or integer that represents name or index of attribute
    :return: Value of attribute
    """
    try:
        return getattr(obj, attr)
    except (TypeError, AttributeError):
        classname = obj.__class__.__name__
        track = tb.format_exc()
        try:
            return obj[attr]
        except KeyError:
            try:
                attr = literal_eval(attr)
                return obj[attr]
            except KeyError:
                track += tb.format_exc()
                msg = f"'{classname}' object has no attribute '{attr}'."
        except TypeError:
            track += tb.format_exc()
            msg = f"Object of type '{classname}' is not subscriptable. Failed to deeply get '{attr}' from it"
    raise dessia_common.errors.DeepAttributeError(message=msg, traceback_=track)


def concatenate_attributes(prefix, suffix, type_: str = 'str'):
    wrong_prefix_format = 'Attribute prefix is wrongly formatted.'
    wrong_prefix_format += 'Is of type {}. Should be str or list'
    if type_ == 'str':
        if isinstance(prefix, str):
            return prefix + '/' + str(suffix)
        if is_sequence(prefix):
            return sequence_to_deepattr(prefix) + '/' + str(suffix)
        raise TypeError(wrong_prefix_format.format(type(prefix)))

    if type_ == 'sequence':
        if isinstance(prefix, str):
            return [prefix, suffix]
        if is_sequence(prefix):
            return prefix + [suffix]
        raise TypeError(wrong_prefix_format.format(type(prefix)))

    wrong_concat_type = 'Type {} for concatenation is not supported.'
    wrong_concat_type += 'Should be "str" or "sequence"'
    raise ValueError(wrong_concat_type.format(type_))


# def deepattr_to_sequence(deepattr: str):
#     sequence = deepattr.split('/')
#     healed_sequence = [a for a in sequence]
#     return healed_sequence


def sequence_to_deepattr(sequence):
    healed_sequence = [str(attr) if isinstance(attr, int) else attr for attr in sequence]
    return '/'.join(healed_sequence)


def type_from_annotation(type_, module):
    """
    Clean up a proposed type if there are stringified
    """
    if isinstance(type_, str):
        # Evaluating types
        type_ = TYPES_FROM_STRING.get(type_, default=getattr(import_module(module), type_))
    return type_


def prettyname(namestr):
    """
    Creates a pretty name from as str
    """
    pretty_name = ''
    if namestr:
        strings = namestr.split('_')
        for i, string in enumerate(strings):
            if len(string) > 1:
                pretty_name += string[0].upper() + string[1:]
            else:
                pretty_name += string
            if i < len(strings) - 1:
                pretty_name += ' '
    return pretty_name


def inspect_arguments(method, merge=False):
    # Find default value and required arguments of class construction
    argspecs = inspect.getfullargspec(method)
    nargs, ndefault_args = split_argspecs(argspecs)

    default_arguments = {}
    arguments = []
    for iargument, argument in enumerate(argspecs.args[1:]):
        if argument not in _FORBIDDEN_ARGNAMES:
            if iargument >= nargs - ndefault_args:
                default_value = argspecs.defaults[ndefault_args - nargs + iargument]
                if merge:
                    arguments.append((argument, default_value))
                else:
                    default_arguments[argument] = default_value
            else:
                arguments.append(argument)
    return arguments, default_arguments


def split_argspecs(argspecs) -> Tuple[int, int]:
    nargs = len(argspecs.args) - 1

    if argspecs.defaults is not None:
        ndefault_args = len(argspecs.defaults)
    else:
        ndefault_args = 0
    return nargs, ndefault_args


def get_attribute_names(object_class):
    attributes = [attribute[0] for attribute in inspect.getmembers(object_class, lambda x:not inspect.isroutine(x))
                  if not attribute[0].startswith('__')
                  and not attribute[0].endswith('__')
                  and isinstance(attribute[1], (float, int, complex, bool))]
    subclass_attributes = {name: param for name, param in inspect.signature(object_class.__init__).parameters.items()
                           if type in inspect.getmro(param.annotation.__class__)}
    subclass_numeric_attributes = [name for name, param in subclass_attributes.items()
                                   if any(item in inspect.getmro(param.annotation)
                                          for item in [float, int, bool, complex])]
    attributes += [attribute for attribute in subclass_numeric_attributes
                   if attribute not in _FORBIDDEN_ARGNAMES]
    return attributes



"""
displays for dessia_common

"""
import warnings
import webbrowser
import os
import tempfile
import inspect
import json


from typing import Union, List
from networkx import DiGraph, Graph, kamada_kawai_layout
from dessia_common.templates import visjs_template
from dessia_common.typings import JsonSerializable
from dessia_common.utils.types import is_sequence


class DisplaySetting:
    def __init__(self, selector, type_, method, arguments=None, serialize_data: bool = False):
        """
        Describe what method to call to get a display
        """
        self.selector = selector
        self.type = type_
        self.method = method
        if not arguments:
            arguments = {}
        self.arguments = arguments
        self.serialize_data = serialize_data

    def to_dict(self):
        """
        Serialization
        """
        return {'selector': self.selector, 'type': self.type, 'method': self.method,
                'serialize_data': self.serialize_data, 'arguments': self.arguments}

    def compose(self, attribute: str, serialize_data: bool = False):
        """
        In case of a parent getting the display settings of a children this methods allow to inject the attribute name
        to method name
        """
        return DisplaySetting(selector=self.selector, type_=self.type, method=f'{attribute}.{self.method}',
                              arguments=self.arguments, serialize_data=serialize_data)


class DisplayObject:
    def __init__(self, type_: str, data: Union[JsonSerializable, List[JsonSerializable], str],
                 reference_path: str = '', traceback: str = '', name: str = ''):
        """
        Container for data of display
        A traceback can be set if display fails to be generated.
        """
        if type_ == "plot_data" and not is_sequence(data):
            warnings.warn("A plot_data DisplayObject must be called with data as a sequence. "
                          "Please return a list of PlotData objects as the result of plot_data method. "
                          "Change have been made automatically", Warning)
            data = [data]
        self.type_ = type_
        self.data = data
        self.traceback = traceback
        self.reference_path = reference_path
        self.name = name

        if data and type_ == 'markdown':
            self.data_cleaning()

    def data_cleaning(self):
        """
        Cleanup tabs in markdown
        """
        self.data = inspect.cleandoc(self.data)

    def to_dict(self):
        """
        Simple serialization
        """
        return {'type_': self.type_, 'data': self.data, 'traceback': self.traceback,
                'reference_path': self.reference_path, 'name': self.name}


def networkx_to_visjs_data(networkx_graph: Graph):
    """
    Compute visjs data to plot from a networkx graph
    """
    visjs_data = {'name': networkx_graph.name, 'nodes': [], 'edges': []}

    pos = kamada_kawai_layout(networkx_graph)

    for i, node in enumerate(networkx_graph.nodes):
        node_dict = networkx_graph.nodes[node]
        node_data = {'id': i}

        if node in pos:
            node_data['x'], node_data['y'] = pos[node]

        if 'name' not in node_dict and 'label' not in node_dict:
            if isinstance(node, str):
                node_data['label'] = node
            elif isinstance(node, int):
                node_data['label'] = str(node)
            else:
                node_data['label'] = ''
            node_data['title'] = node_data['label']
            if len(node_data['label']) > 10:
                node_data['label'] = '[ ]'
        elif 'name' in node_dict and 'label' not in node_dict:
            node_data['label'] = node_dict['name']
        else:
            node_data['label'] = node_dict['name'] + node_dict['label']

        if 'shape' not in node_dict:
            node_data['shape'] = 'circular'
        else:
            node_data['shape'] = node_dict['shape']

        if 'color' in node_dict:
            node_data['color'] = node_dict['color']

        visjs_data['nodes'].append(node_data)

    list_nodes = list(networkx_graph.nodes)
    is_digraph = isinstance(networkx_graph, DiGraph)
    # print(is_digraph)
    for edge in networkx_graph.edges:
        index1 = list_nodes.index(edge[0])
        index2 = list_nodes.index(edge[1])
        edge_nx_data = networkx_graph.get_edge_data(*edge)

        edge_data = {'from': index1,
                     'to': index2,
                     'font': {'align': 'middle'}}

        if is_digraph:
            if 'head_type' in edge_nx_data:
                edge_data['arrows'] = {'to': {'enabled': True, 'type': edge_nx_data['head_type']}}
            else:
                edge_data['arrows'] = 'to'

        if 'color' in edge_nx_data:
            edge_data['color'] = {'color': edge_nx_data['color']}

        visjs_data['edges'].append(edge_data)

    return visjs_data


def draw_networkx_graph(networkx_graph: Graph):
    """
    Draw a networkx graph in a browser using VisJS library
    """
    visjs_data = networkx_to_visjs_data(networkx_graph)
    content = visjs_template.substitute(nodes=json.dumps(visjs_data['nodes']),
                                        edges=json.dumps(visjs_data['edges']),
                                        name=visjs_data['name'])
    with tempfile.NamedTemporaryFile(suffix=".html",
                                     delete=False) as file:
        file.write(bytes(content, 'utf8'))

    # with open('graph_visJS.html', 'wb') as file:
    #     file.write(s.encode('utf-8'))
    webbrowser.open('file://' + os.path.realpath(file.name))
    return file.name
